/*  Game Input
    Copyright (C) 2001  Clinton Ebadi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


/* TODO:
 * add throw() to methods that cannot throw things
 * add throw lists for the exceptions a function can throw
 * create exceptions for game_handler_funcs
 * I/O for game_handler_funcs
 */
#ifndef CKE_GAME_INPUT_HH
#define CKE_GAME_INPUT_HH

#include <vector>
#include <string>
#include <map>
#include <iostream>

// use str_vec in callback prototypes
typedef std::vector<std::string> str_vec;

// event handling function pointer, the vector is a set of
//  strings that represent its  arguments
class game_parser; // for next typedef
typedef bool (*game_handler_func) (str_vec, game_parser&);

// game_handler type
struct game_handler_t {
  game_handler_t (game_handler_func f=0, std::string m="", unsigned int u=0, int i=0)
    : func (f), args_msg (m), req_args (u), opt_args (i) { }
  game_handler_func func;
  std::string args_msg; // message given when wrong number of args
  unsigned int req_args; // required arguments
  int opt_args;
};

// explanation of game_handler_t::opt_args
// 0 is no optional args, any positive number lists a
// number of optional args. The user may specify up to
// that many optional args. Any more triggers an error.
// -1 signifies that you have an unlimited number of
// arguments. Any other negative numbers are ignored,
// and treated as if they were 0

class game_parser { // game input parser
public:
  game_parser ();

  // event handler functions

  // takes a key and a game_handler_t
  // game_handler_t contains a callback and related information
  // is encountered
  // note: can be used to change a handler as well as add one
  void event_handler (std::string, game_handler_t);

  // removes handler
  void remove_event_handler (const std::string);


  // input functions

  // associated object with input stream
  // defaults to std::cin
  void input_stream (std::istream*);
  void output_stream (std::ostream*); // sets ostream (I know this is the input area)

  // sets command delimeter
  // returns old one (default is \n)
  char command_delimeter (char);

  // gets the next command
  // throws an exception on internal failure
  // returns true if no warnings or errors occured in handler
  bool get_command ();

  // exceptions that may be thrown
  // these should only be thrown by the game_parser

  // thrown when not enough args are given to a keyword
  struct not_enough_args {
    not_enough_args (std::string, std::string, unsigned int, unsigned int, unsigned int);
    std::string cmd; // cmd called
    std::string msg; // optional message to display
    unsigned int given; // args given
    unsigned int min; // min args
    unsigned int max; // max args
  };

  // key not found in map
  struct key_not_found {
    key_not_found (std::string);
    std::string key;
  };

  // methods/types for usage by game_handler_funcs
  // used to display messages/errors

  std::ostream* get_output_stream () { return output; } // returns the output stream
  std::istream* get_input_stream () { return input; } // returns current input stream

  // throw this on error/warning
  struct handler_error { // handler error
    handler_error (std::string m, bool t)
      : message(m), fatal (t)
    { }
    std::string message; // message to print
    bool fatal; // is this fatal? (return false?)
  };


private:

  // convenience typedef
  typedef std::map<std::string, game_handler_t> hmap_t;

  hmap_t handlers; // handlers
  std::istream* input; // input reference
  std::ostream* output; // output reference
  char c_del;

};

#endif
